/*
 *  html.c
 *  mod_musicindex
 *
 *  $Id: html.c,v 1.29 2003/06/25 15:04:37 varenet Exp $
 *
 *  Created by Thibaut VARENE on Thu Mar 20 2003.
 *  Copyright (c) 2003 Regis BOUDIN
 *  Copyright (c) 2003 Thibaut VARENE
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1, or (at your option)
 *  any later version.
 *
 */

/**
 * @file 
 * HTML I/O handling.
 *
 * That file groups all functions related to HTML generation
 * and parsing, that is to say functions that are used to
 * send HTML code to the client and functions that are in
 * the process of parsing URI requests.
 * Here are also the functions related to m3u playlist creation.
 */
 
#include "html.h"
#include "inf.h"

/**
 * Global string used in URL processing.
 *
 * That global variable is mandatory since it's used by ascii2rq to
 * return a proper request string.
 */
char chaine[MAX_STRING];

/**
 * Escapes special chars in an ascii string to make it request compliant.
 * 
 * @param orig The string to escape.
 *
 * @return The escaped string.
 */
static char *ascii2rq(char *orig)
{
	char a;
	int i=0, j=0;
	
	while ((a=orig[i]) != '\0') {
		switch (a) {
			case ' ' : chaine[j]='+'; break;
			case '%' : chaine[j++]='%'; chaine[j++]='2'; chaine[j]='5'; break;
			case ';' : chaine[j++]='%'; chaine[j++]='3'; chaine[j]='B'; break;
			case '?' : chaine[j++]='%'; chaine[j++]='3'; chaine[j]='F'; break;
			case ':' : chaine[j++]='%'; chaine[j++]='3'; chaine[j]='A'; break;
			case '@' : chaine[j++]='%'; chaine[j++]='4'; chaine[j]='0'; break;
			case '&' : chaine[j++]='%'; chaine[j++]='2'; chaine[j]='6'; break;
			case '=' : chaine[j++]='%'; chaine[j++]='3'; chaine[j]='D'; break;
			case '+' : chaine[j++]='%'; chaine[j++]='2'; chaine[j]='B'; break;
			case '$' : chaine[j++]='%'; chaine[j++]='2'; chaine[j]='4'; break;
			case '<' : chaine[j++]='%'; chaine[j++]='3'; chaine[j]='C'; break;
			case '>' : chaine[j++]='%'; chaine[j++]='3'; chaine[j]='E'; break;
			case '#' : chaine[j++]='%'; chaine[j++]='2'; chaine[j]='3'; break;
			case '"' : chaine[j++]='%'; chaine[j++]='2'; chaine[j]='2'; break;
			case '{' : chaine[j++]='%'; chaine[j++]='7'; chaine[j]='B'; break;
			case '}' : chaine[j++]='%'; chaine[j++]='7'; chaine[j]='D'; break;
			case '|' : chaine[j++]='%'; chaine[j++]='7'; chaine[j]='C'; break;
			case '\\' : chaine[j++]='%'; chaine[j++]='5'; chaine[j]='C'; break;
			case '^' : chaine[j++]='%'; chaine[j++]='5'; chaine[j]='E'; break;
			case '[' : chaine[j++]='%'; chaine[j++]='5'; chaine[j]='B'; break;
			case ']' : chaine[j++]='%'; chaine[j++]='5'; chaine[j]='D'; break;
			case '`' : chaine[j++]='%'; chaine[j++]='6'; chaine[j]='0'; break;
			default : chaine[j] = a;
		}
		i++; j++;
	}
	chaine[j] = '\0';
	return chaine;
}

/**
 * Formats the list of songs to be sent.
 *
 * This function formats a list of tracks (if any) to be sent
 * to the client, as found in the p list passed in argument.
 * Depending on config options, it allows streaming/shuffle etc.
 * 
 * @param r Apache request_rec struct to handle connection details.
 * @param p struct mu_ent.
 * @param conf MusicIndex configuration paramaters struct.
 */
static void list_songs(request_rec * r, struct mu_ent *p, mu_config * conf)
{
	struct mu_ent *q = p;
	unsigned short i = 0, j = 1;
	
	while (q) {
		if (q->filetype != FT_DIR)
			break;
		q = q->next;
	}

	if (!q)
		return;

	while (q) {
		if (j) {
			ap_rputs("<tr class=\"normal\"><td>", r);
			j = 0;
		} else {
			ap_rputs("<tr class=\"highlight\"><td>", r);
			j = 1;
		}
	
		if (conf->options & (MI_ALLOWDWNLD | MI_ALLOWSTREAM)) {
			ap_rvputs(r, "<input type=\"checkbox\" name=\"file\" value=\"",
				q->file, "\" />", NULL);
		
			if (conf->options & MI_ALLOWDWNLD)	/*Display [fetch] */
				ap_rvputs(r, "&nbsp;<a href=\"", ap_escape_uri(r->pool, q->file), "\">"
					"<img alt=\"\" border=\"0\" src=\"", conf->fetch_icon, "\"></a>", NULL);
		
			if (conf->options & MI_ALLOWSTREAM)	/*Display [stream] */
				ap_rvputs(r, "&nbsp;<a href=\"?file=", ascii2rq(q->file), "&amp;action=playlist\">"
					"<img alt=\"\" border=\"0\" src=\"", conf->sound_icon, "\"></a>", NULL);
		}
		
		ap_rputs("</td>", r);
			
		for (i = 0; conf->fields[i] != '\0'; i++) {
			switch (conf->fields[i]) {
				case SB_TITLE:
					ap_rvputs(r, "<td>", q->title, "</td>", NULL);
					break;
				case SB_TRACK:
					if (q->track)
						ap_rprintf(r, "<td align=\"center\">%d</td>", q->track);
					else
						ap_rputs( "<td></td>", r);
					break;
				case SB_ARTIST:
					ap_rvputs(r, "<td>", q->artist, "</td>", NULL);
					break;
				case SB_LENGTH:
					if (q->length)
						ap_rprintf(r, "<td>%ld:%.2ld</td>", q->length / 60, q->length % 60);
					else
						ap_rputs( "<td></td>", r);
					break;
				case SB_BITRATE:
					if (q->bitrate)
						ap_rprintf(r, "<td align=\"RIGHT\">%ld&nbsp</td>", q->bitrate >> 10);
					else
						ap_rputs( "<td></td>", r);
					break;
				case SB_DATE:
					if (q->date)
						ap_rprintf(r, "<td align=\"RIGHT\">%d&nbsp</td>", q->date);
					else
						ap_rputs( "<td></td>", r);
					break;
				case SB_ALBUM:
					ap_rvputs(r, "<td>", q->album, "</td>", NULL);
					break;
				case SB_FILETYPE:
					switch(q->filetype) {
						case FT_MP3:
							ap_rvputs(r, "<td>MP3</td>", NULL);
							break;
						case FT_OGG:
							ap_rvputs(r, "<td>OGG</td>", NULL);
							break;
					}
					break;
			}
		}
		ap_rputs("</tr>\n", r);
		q = q->next;
	}

}

/**
 * Splits song list by folders.
 *
 * This function takes a mu_ent list of tracks as arguments and
 * formats it nicely to differenciate subfolders.
 * 
 * @param r Apache request_rec struct to handle connection details.
 * @param p struct mu_ent.
 * @param conf MusicIndex configuration paramaters struct.
 */
static void split_folders(request_rec * r, struct mu_ent *p, mu_config * conf)
{
	struct mu_ent *intermed = NULL, *temp = p;
	char *tembuf;

	if (!p)
		return;

	tembuf = ap_make_dirstr_parent(r->pool, p->file);
		
	while (temp->next && !strcmp(tembuf, ap_make_dirstr_parent(r->pool, temp->next->file)))
		temp = temp->next;

	intermed = temp->next;
	temp->next = NULL;
		
	if (strlen(tembuf) > 1)		/* We're out of the current folder */
		ap_rvputs(r, "<tr class=\"title\"><th align=\"left\" colspan=\"10\">in <a href=\"", tembuf, "\">", tembuf, "</a></th>\n", NULL);
	else
		ap_rputs("<tr class=\"title\"><th align=\"left\" colspan=\"10\">in Current Directory</th>\n", r);
				
	list_songs(r, p, conf);
	
	/* reentrant call for next subfolder */
	if (intermed != NULL)
		split_folders(r, intermed, conf);
	
	return;
}

/**
 * Treats the URL arguments.
 *
 * This function searches for keywords passed as URL arguments (with "?xxx")
 * and sets the handler options accordingly.
 * 
 * @param r Apache request_rec struct to handle connection details.
 * @param conf MusicIndex configuration paramaters struct.
 *
 * @return The corresponding option letter.
 */
char treat_args(request_rec * r, mu_config * conf)
{
	const char *s = r->args;
	const char *p;
	unsigned short i;

	if ((s == NULL) || (*s == '\0'))
		return 'A';

	while (s[0]){
		p = ap_getword(r->pool, &s, '&');
		if (!strncmp(p, "action=", 7)) {
			p += 7;
			if ((!strcmp(p, "playlist") || !strcmp(p, "Play+Selected")) && (conf->options & MI_ALLOWSTREAM)) {
				return 'L';
			} else if ((!strcmp(p, "playall") || !strcmp(p, "Play+All")) && (conf->options & MI_ALLOWSTREAM)) {
				return 'P';
			} else if (!strcmp(p, "Shuffle+All") && (conf->options & MI_ALLOWSTREAM)) {
				conf->order[0] = SB_RANDOM;
				conf->order[1] = SB_DEFAULT;
				set_fctptrs(conf);
				return 'P';
			} else if (!strcmp(p, "Search") && (conf->options & MI_ALLOWSEARCH)) {
				conf->options |= MI_SEARCH;
			} else if (!strcmp(p, "Recursive+Search") && (conf->options & MI_ALLOWSEARCH)) {
				conf->play_recursive |= MI_RECURSIVE;
			}
				
		} else if (!strncmp(p, "sort=", 5)) {
			p += 5;
			for (i = ARG_NUMBER-1; i!=0; i--)
				conf->order[i] = conf->order[i-1];
			conf->order[0] = p[0];
			set_fctptrs(conf);
		} else if (!strncmp(p, "option=", 7)) {
			p += 7;
			if (!strcmp(p, "recursive")) {
				conf->play_recursive |= MI_RECURSIVE;
			} else if (!strcmp(p, "shuffle")) {
				conf->order[0] = SB_RANDOM;
				conf->order[1] = SB_DEFAULT;
				set_fctptrs(conf);
			} else if (!strcmp(p, "quick")) {
				conf->options |= MI_QUICKPL;
			} else {
				conf->search = ap_pstrdup(r->pool, p);
				for (i=0; p[i]; i++) {
					if (p[i] == '+') {
						conf->search[i] = ' ';
					}
				}
				ap_unescape_url(conf->search);
			}
		}
	}

	return 'A';
}

/**
 * Formats a complete url string to be sent to client.
 *
 * This function does all the URL formatting, in the form:
 * http://[user:][passwd\@]hostname[:port]/uri/
 * 
 * @param r Apache request_rec struct to handle connection details.
 * @param uri A string (basically a path) to be sent to form a complete url.
 * @param command An eventual command (like ?action=playall).
 */
void send_url(request_rec *r, char *uri, char *command)
{
	char *prefix = malloc(MAX_PREFIX);
	char str_port[6];
	char *bufcoded, *decoded;
	unsigned short l;
	
	strcpy(prefix, "http://");
	
	if (r->connection->user) {
		/* grab the auth credentials base64 encoded */
		const char *auth = ap_table_get(r->headers_in, "Authorization");
		if (auth) {
			bufcoded = strrchr(auth, ' ');
			bufcoded++;
			decoded = (char *)ap_palloc(r->pool, 1 + ap_base64decode_len(bufcoded));
			l = ap_base64decode(decoded, bufcoded);
			decoded[l] = '\0'; /* make binary sequence into string */
			strcat(prefix, decoded);
		}
		strcat(prefix, "@");
	}

	strcat(prefix, r->hostname);
	
	if (r->server->port != 80) {
		sprintf(str_port, "%u", r->server->port);
		strcat(prefix, ":");
		strcat(prefix, str_port);
	}
	
	ap_rvputs(r, prefix, ap_escape_uri(r->pool, uri), NULL);
	if (command)
		ap_rvputs(r, command, NULL);
	ap_rvputs(r, "\n", NULL);
	
	free(prefix);
}	
		
/**
 * Sends HTML page headers and top of the page.
 *
 * This function takes care of the formating of the HTML page headers.
 * It sends the data that is common to all pages generated by the handler.
 * Currently it relies on apache_mp3 original stylesheets.
 * 
 * @param r Apache request_rec struct to handle connection details.
 * @param conf MusicIndex configuration paramaters struct.
 */
void send_head(request_rec * r, mu_config * conf)
{
	char *s, *t, *u;
	FILE *in = NULL;

	s = ap_pstrdup(r->pool, r->uri);

	ap_rvputs(r, "<!DOCTYPE HTML PUBLIC \"-//W3C//"
		"DTD HTML 4.01 Transitional//EN\"\n"
		"\"http://www.w3.org/TR/REC-html4/loose.dtd\">\n"
		"<HTML>\n<HEAD>\n<META http-equiv=\"Content-Type\""
		"content=\"text/html; charset=UTF-8\">\n"
		"<link rel=\"stylesheet\" type=\"text/css\" "
		"href=\"", conf->css, "\">\n", NULL);

	ap_rvputs(r, "<TITLE>Musical index of ", r->uri,
		"</TITLE>\n</HEAD>\n<BODY>\n", NULL);

	u = s + strlen(s);
	if (strlen(s) > 1)
		u--;
	if (*u == '/')
		*u = '\0';
	while (*u != '/')
		u--;
	u++;

	ap_rputs("<img align=\"LEFT\" alt=\"Dir\" border=\"0\" ", r);

	if ((in = fopen(ap_pstrcat(r->pool, r->filename, "/", "cover.png", NULL), "r"))) {
		ap_rputs("src=\"cover.png\">\n", r);
		fclose(in);
	}
	else if ((in = fopen(ap_pstrcat(r->pool, r->filename, "/", "cover.jpg", NULL), "r"))) {
		ap_rputs("src=\"cover.jpg\">\n", r);
		fclose(in);
	}
	else {
		ap_rvputs(r, "src=\"", conf->cd_icon, "\">\n", NULL);
	}

	t = u = s + 1;

	if (*u != '\0')
		ap_rputs("<h1>&nbsp;<a href=\"/\">Music</a>", r);
	else
		ap_rputs("<h1>&nbsp;Music", r);

	do {
		while ((*u != '/') && (*u != '\0'))
			u++;

		if (u != s + 1)
			ap_rvputs(r, "&nbsp;<img src=\"", conf->arrow,
				"\" alt=\"=>\">&nbsp;", NULL);

		*u = '\0';
		ap_rvputs(r, "<a href=\"", s, "/\">", t, "</a>\n", NULL);
		*u = '/';
		u++;
		t = u;

	} while (*u != '\0');

	ap_rputs("</h1>", r);

	if (conf->options & MI_ALLOWSTREAM)
		ap_rputs("<a href=\"?option=recursive&amp;option=shuffle&amp;action=playall\"><font "
			"class=\"directory\">[Shuffle All]</font></a>&nbsp;"
			"<a href=\"?option=recursive&amp;action=playall\"><font class=\"directory\">"
			"[Stream All]</font></a>\n", r);

	/* simply displays a searching box if option is activated */
	if (conf->options & MI_ALLOWSEARCH){
		ap_rvputs(r, "<br><br><form method=\"get\" action=\"", ap_escape_uri(r->pool, r->uri), "\""
		"enctype=\"application/x-www-form-urlencoded\" name=\"searching\">",
		"<input type=\"text\" name=\"option\">",
		"<input type=\"hidden\" name=\"action\" value=\"Search\">",
		"<input type=\"submit\" name=\"action\" value=\"Search\">",
		"<input type=\"submit\" name=\"action\" value=\"Recursive Search\">",
		"</form>\n\n", NULL);
	}

	ap_rputs("<hr>\n<!-- begin main -->\n\n", r);

}

/**
 * Sends directory listing for the current folder.
 *
 * This function takes care of preparing and sending to the client the
 * list (if any) of the available directories in the current folder.
 * Depending on config options, it allows streaming/shuffle etc.
 * 
 * @param r Apache request_rec struct to handle connection details.
 * @param p struct mu_ent.
 * @param conf MusicIndex configuration paramaters struct.
 */
void send_directories(request_rec * r, struct mu_ent * p, mu_config * conf)
{
	struct mu_ent *q = p;
	unsigned short i = 0, nb = 0;
	char *c;
	char temp[MAX_STRING];

	if (!q)
		return;

	if (q->filetype != FT_DIR)
		return;
	
	/* We count the number of directories for later use */
	while (q && (q->filetype == FT_DIR)) {
		nb++;
		q = q->next;
	}
	
	q = p;
		
	ap_rprintf(r, "<!-- begin subdirs -->\n<h2 class=\"CDdirectories\">Music Directories (%d)</h2>\n\n", nb);

	ap_rputs("<center><table width=\"95%\" border=\"0\">\n", r);

	i = 1;
	while (q && (q->filetype == FT_DIR)) {

		c = ap_cpystrn(temp, q->file, MAX_STRING)-1;
		*c = '\0';

		if (i == 1)
			ap_rputs("<tr valign=\"BOTTOM\">", r);

		ap_rvputs(r, "<td align=\"left\"><a href=\"", ap_escape_uri(r->pool, q->file), NULL);

		if (conf->options & MI_ALLOWSTREAM)
			ap_rputs("?option=recursive&amp;action=playall", r);

		ap_rvputs(r, "\"><img class=\"subdir\" align=\"MIDDLE\" "
			"alt=\"\" border=\"0\" "
			"src=\"", conf->small_cd_icon, "\" /></a>&nbsp;", NULL);

		ap_rvputs(r, "<a href=\"", ap_escape_uri(r->pool, q->file), "\">",
	    		"<font class=\"subdirectory\">", temp, "</font></a>",
	    		"<br>", NULL);

		if (conf->options & MI_ALLOWSTREAM) {
			ap_rvputs(r, "<a class=\"subdirbuttons\" href=\"",
				ap_escape_uri(r->pool, q->file), "?option=recursive&amp;option=shuffle&amp;option=playall\">"
				"[Shuffle]</a>&nbsp;", NULL);

			ap_rvputs(r, "<a class=\"subdirbuttons\" href=\"",
				ap_escape_uri(r->pool, q->file), "?option=recursive&amp;action=playall\">"
				"[Stream]</a></td>", NULL);
		}

		if (i == 3) {
			i = 0;
			ap_rputs("</tr>\n", r);
		} else
			ap_rputs("\n", r);
		i++;
		q = q->next;
	}

	ap_rputs("</table></center><br><hr>\n<!-- end subdirs -->\n\n", r);
}

/**
 * Sends track listing for the current folder.
 *
 * This function takes care of preparing and sending to the client the
 * list (if any) of the available songs in the current p list.
 * Depending on config options, it allows streaming/shuffle etc.
 * 
 * @param r Apache request_rec struct to handle connection details.
 * @param p struct mu_ent.
 * @param conf MusicIndex configuration paramaters struct.
 */
void send_tracks(request_rec * r, struct mu_ent *p, mu_config * conf)
{
	struct mu_ent *q = p;
	unsigned short i = 0, nb = 0;
	
	while (q) {
		if (q->filetype != FT_DIR)
			break;
		q = q->next;
	}

	if (!q)
		return;

	/* We count the number of songs for later use */
	while (q) {
		if (q->filetype != FT_DIR)
			nb++;
		q = q->next;
	}

	ap_rputs("<!-- begin tracks -->\n<h2 class=\"SongList\">", r);
	
	ap_rvputs(r, "<form method=\"get\" action=\"", ap_escape_uri(r->pool, r->uri), "\" "
		"enctype=\"application/x-www-form-urlencoded\" name=\"form\">\n\n", NULL);
	
	if (conf->options & MI_SEARCH)
		ap_rputs("Result", r);
	else
		ap_rputs("Song", r);
	
	ap_rprintf(r, "	List (%d)</h2>\n\n", nb);

	ap_rputs("<table width=\"100%\" cellspacing=\"0\" border=\"0\">", r);

	if (conf->options & (MI_ALLOWDWNLD | MI_ALLOWSTREAM))
		ap_rputs("<tr align=\"LEFT\" class=\"title\">"
			"<th align=\"CENTER\">Select</th>", r);
	else
		ap_rputs("<tr align=\"LEFT\" class=\"title\">"
			"<th align=\"CENTER\"></th>", r);
	
	for (i = 0; conf->fields[i] != '\0'; i++) {	/*Diplay title line*/
		switch (conf->fields[i]) { 
			case SB_TITLE:
				ap_rprintf(r, "<th><a href=\"?sort=%c\">Title</a></th>", SB_TITLE);
				break;
			case SB_TRACK:
				ap_rprintf(r, "<th><a href=\"?sort=%c\">Track</a></th>", SB_TRACK);
				break;
			case SB_ARTIST:
				ap_rprintf(r, "<th><a href=\"?sort=%c\">Artist</a></th>", SB_ARTIST);
				break;
			case SB_LENGTH:
				ap_rprintf(r, "<th><a href=\"?sort=%c\">Duration</a></th>", SB_LENGTH);
				break;
			case SB_BITRATE:
				ap_rprintf(r, "<th><a href=\"?sort=%c\">Bitrate</a></th>", SB_BITRATE);
				break;
			case SB_ALBUM:
				ap_rprintf(r, "<th><a href=\"?sort=%c\">Album</a></th>", SB_ALBUM);
				break;
			case SB_DATE:
				ap_rprintf(r, "<th><a href=\"?sort=%c\">Date</a></th>", SB_DATE);
				break;
			case SB_FILETYPE:
				ap_rprintf(r, "<th><a href=\"?sort=%c\">Type</a></th>", SB_FILETYPE);
				break;
		}
	}
	ap_rputs("</tr>\n", r);

	if ((conf->options & MI_SEARCH) && (conf->play_recursive & MI_RECURSIVE))
		split_folders(r, p, conf);
	else
		list_songs(r, p, conf);

	ap_rvputs(r, "<tr><td align=\"left\" colspan=\"10\">"
/*		"<!--<input type=\"submit\" name=\"Add to Playlist\" value=\"Add to Playlist\" class=\"playlist\">-->"
		"<!--<input type=\"submit\" name=\"Add All to Playlist\" value=\"Add All to Playlist\" class=\"playlist\">-->" */
		"<input type=\"submit\" name=\"action\" value=\"Play Selected\">"
		"<input type=\"submit\" name=\"action\" value=\"Shuffle All\">"
		"<input type=\"submit\" name=\"action\" value=\"Play All\">"
		"</td></tr>\n</table>\n"
/*		"<!--<input type=\"hidden\" name=\".cgifields\" value=\"playlist\">-->\n" */
		"</form><br><hr>\n<!-- end tracks -->\n\n", NULL);
}

/**
 * Sends playlist with url and EXTM3U info to the client.
 *
 * This function prepares an m3u playlist (with extra information where
 * available: length, artist, title, album) and sends it to the client.
 * 
 * @todo We have to send it with ".m3u" file extension othewise it's not recognized.
 * @param r Apache request_rec struct to handle connection details.
 * @param p struct mu_ent.
 * @param conf MusicIndex configuration paramaters struct.
 */
void send_playlist(request_rec * r, struct mu_ent *p, mu_config * conf)
{
	struct mu_ent *q = p;
	if (!p)
		return;

	ap_rputs("#EXTM3U\n", r);

	while (q)
	{
		ap_rprintf(r, "#EXTINF:%li,", q->length);
		if (q->artist)
			ap_rvputs(r, q->artist, " - ", NULL);
		ap_rvputs(r, q->title, NULL);
		if (q->album)
			ap_rvputs(r, " (", q->album, ")", NULL);
		ap_rvputs(r, "\n", NULL);
		send_url(r, q->uri, NULL);
		q = q->next;
	}
}

/**
 * Sends HTML page footers.
 *
 * This function takes care of the formating of the HTML page footers.
 * It sends the data that is common to all pages generated by the handler.
 * Currently it relies on apache_mp3 original stylesheets.
 * 
 * @param r Apache request_rec struct to handle connection details.
 * @param conf MusicIndex configuration paramaters struct.
 */
void send_foot(request_rec * r, mu_config * conf)
{
	ap_rprintf(r, "<!-- end main -->\n<!-- mod_musicindex v.%s -->\n", MUSIC_VERSION_STRING);
	ap_rprintf(r, "<!-- Authors: %s -->\n", MUSIC_AUTHORS_STRING);
	ap_rputs("<table border=\"0\" width=\"100%\"><tr>"
		"<td align=\"left\"><p><a href=\"http://validator.w3.org/check/referer\">"
		"<img border=\"0\"\nsrc=\"http://www.w3.org/Icons/valid-html401\"\n"
        	"alt=\"Valid HTML 4.01!\" height=\"31\" width=\"88\"></a></p></td>\n", r);
	ap_rprintf(r, "<td align=\"right\">MusicIndex v.%s</td>\n", MUSIC_VERSION_STRING);
	ap_rputs("</tr></table>\n</BODY>\n</HTML>", r);
}
